/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include <pthread.h>
#include <iostream>
#include <sstream>  
#include <ctime>
#include <algorithm>    // std::sort
#include <vector> 
#include "Geometry.h"
#include "Transform.h"

#include "Triangle.h"

#include "Mesh.h"
#include "Image.h"

#include "MeshIO.h"
#include "ImageIO.h"
#include "tree.h"

#define NB_THREAD 8

#define NB_DIRECT 10
#define NB_INDIRECT 100
#define INDIRECT_DEPTH 2
#define MIRROR_DEPTH 2

#define EPSILON_MIRROR 0.003
#define X 512
#define Y 512

#define EXTEND_MOD 3
#define RAND_LIGHT 2


BSP m_bsp;
double sourcesArea = 0;

gk::Color trace(const gk::Ray& ray, int depth);

typedef struct
{
  gk::Color* c;
  gk::Image *image;
  gk::Transform vpv;
  pthread_t thread_render;
  pthread_t thread_renders [NB_THREAD];
 
  //   pthread_mutex_t mutex_stock;
} render_t;
 
typedef struct str_thdata
{
  int thread_no;
} thdata;

static render_t render; // TODO : variable global => bof


// objet
gk::Mesh *mesh= NULL;

// representation d'une source de lumiere
struct Source
{
  gk::Triangle triangle;
  gk::Energy emission;
};

double frand(double fMin, double fMax)
{
  double f = (double)rand() / RAND_MAX;
  return fMin + f * (fMax - fMin);
}

int irand(int rangeLow, int rangeHigh) {
  return rangeLow+rand()%(rangeHigh-rangeLow);
}


gk::Color Normalize(const gk::Color& e){
  gk::Vector result(e.r,e.g,e.b);
  if(result.Length())
    result = Normalize(result);
  return gk::Color(result.x,result.y,result.z,e.a);
}


gk::Color Max(const gk::Color& e,float v){
  gk::Vector result(std::min(e.r,v),std::min(e.g,v),std::min(e.b,v));
  return gk::Color(result.x,result.y,result.z,e.a);
}

bool compareSources(const Source& a, const Source& b){
  return a.triangle.area()> b.triangle.area();
}

// ensemble de sources de lumieres
std::vector<Source> sources;
std::vector<gk::MeshMaterial> materials;



// recuperer les sources de lumiere du mesh
int build_sources( const gk::Mesh *mesh )
{
  for(int i= 0; i < mesh->triangleCount(); i++)
    {
      // matiere associe a chaque triangle de l'objet
      const gk::MeshMaterial& material= mesh->triangleMaterial(i);
       
      if(gk::Color(material.emission).isBlack() == false)
        {
	  // construire une source de lumiere
	  Source source;
	  source.emission= material.emission;
	  source.triangle= mesh->triangle(i);
	  sourcesArea +=source.triangle.area();
	  // inserer la source de lumiere dans l'ensemble.
	  sources.push_back(source);
	  
        }
    }
    
  printf("%lu sources.\n", sources.size());
  printf("surface de lumiere : %lf.\n",sourcesArea);
  return sources.size();
}


// ensemble de triangles
std::vector<gk::Triangle> triangles;

// recuperer les triangles du mesh
int build_triangles( const gk::Mesh *mesh )
{
  for(int i= 0; i < mesh->triangleCount(); i++)
    {
      gk::Triangle triangle= mesh->triangle(i);
      triangles.push_back( triangle );
    }
    
  printf("%lu triangles.\n", triangles.size());
  return triangles.size();
}


// calcule l'intersection d'un rayon et de tous les triangles
bool intersect( const gk::Ray& ray, gk::Hit& hit )
{
  for(unsigned int i= 0; i < triangles.size(); i++)
    {
      float t, u, v;
      if(triangles[i].Intersect(ray, hit.t, t, u, v))
	{
	  hit.t= t;
	  hit.u= u;
	  hit.v= v;
	  hit.p= ray(t);      // evalue la positon du point d'intersection sur le rayon
	  hit.object_id= triangles[i].id;	// permet de retrouver toutes les infos associees au triangle
	}
    }

  return (hit.object_id != -1);
}



gk::Color direct(gk::Point p, gk::Normal n, const gk::MeshMaterial& material){

  unsigned int i, N;
  gk::Color result(0,0,0,0);
  float pdf;
  gk::Point q;
  gk::Color C(material.diffuse_color);


  for(N=0;N<NB_DIRECT;N++){    
#if RAND_LIGHT==1
    i = irand(0, sources.size());
#else
    double tmpArea = irand(0, sourcesArea);
    i=0;
      do{
      tmpArea -= sources[i].triangle.area();
      i++;
      }while(tmpArea>0);
      i--;
#endif
    gk::Color Li = sources[i].emission; 
    gk::Color tmp(0,0,0,0);
    
    pdf = sources[i].triangle.sampleUniform(frand(0,1),frand(0,1),q);
#if EXTEND_MOD==2
    pdf = 1.f/sourcesArea;
#elif EXTEND_MOD==3
    pdf = sourcesArea;
#endif
    gk::Ray ray(p,q);
    gk::Hit hit(ray);
    m_bsp.intersect(ray,hit);
    //intersect(ray, hit);
    if(hit.object_id == -1){
      float pq=Dot(gk::Normalize(q-p),n);
      float qp=Dot(gk::Normalize(p-q),sources[i].triangle.normal());
      float cosThetaIn = std::max( pq, 0.f);
      float cosThetaOut = std::max( qp, 0.f);
      float G = (cosThetaIn*cosThetaOut) / gk::Dot(ray.d,ray.d);
      float fr = material.kd + material.ks * cosThetaIn;
      //float fr =  material.kd  / M_PI;

      tmp += (Li * C * fr * G)* pdf;

      //tmp += (Li * C  * G ) * (float)sources[i].triangle.area();
    }
    
    // tmp *= (float)sources[i].triangle.area();
  
    result += tmp;
  }

  result /= (float)NB_DIRECT  ;
  return result;
}


gk::Color indirect(gk::Point p, gk::Normal n, const gk::MeshMaterial& material, int id, int depth){
  gk::Color result;
  if(depth>=INDIRECT_DEPTH)
    return result;
  

  for(int i = 0; i < NB_INDIRECT; i++){

    gk::Ray ray;
    gk::Hit hit;

    double r1= frand(0,1);
    double r2= frand(0,1);

    double phi = 2.f * M_PI * r1;
    double sqrtR2 = sqrt(1.f - r2 * r2 );
    double theta = acos( r2 );
  
    double x = cos(phi) * sqrtR2;

    double y = sin(phi) * sqrtR2; 

    double z = r2;
    gk::Vector dir = gk::Vector(x,y,z);  
    
    float pdf =  1.f/(2.f*M_PI);
    //    float pdf = 1.f/(2.f*M_PI * (1.f - cos))
    dir = mesh->triangle(id).world(dir); //replacer la direction dans le repère du monde
    ray = gk::Ray(p,dir);
    hit = gk::Hit(ray);
    //    intersect(ray, hit);
    m_bsp.intersect(ray, hit);

    if(hit.object_id != -1){
      gk::Point pTmp= ray(hit.t);
      gk::Normal normalTmp= mesh->triangle(hit.object_id).normal();
      const gk::MeshMaterial& materialTmp= mesh->triangleMaterial(hit.object_id);
      gk::MeshMaterial mUsed(materialTmp);
      //      float pq=Dot(gk::Normalize(pTmp-p),n);
      float cosThetaIn = std::max(0.f, Dot(gk::Vector(n),gk::Normalize(ray.d)) );//std::max( pq, 0.f);
      float fr = material.kd + material.ks * cosThetaIn;
      result = direct(pTmp, normalTmp, mUsed);
      result *=  fr;   
      result += indirect( pTmp, normalTmp, mUsed, hit.object_id, depth+1 ) * pdf;
    }

  }
  //    result += calcul_indirect( p, n, material, NB_REC  , id);

  result/=(double)NB_INDIRECT;
  return result;
}

gk::Color reflexion(const gk::Ray& r, gk::Point p, gk::Normal n, const gk::MeshMaterial& material, int depth){
  if(depth>=MIRROR_DEPTH)
    return gk::Color();
  gk::Vector reflectDirection = r.d - 2*(r.d * gk::Vector(n)) * gk::Vector(n);
  gk::Ray ray(p,reflectDirection);
  return trace(ray, depth + 1);
  
}


gk::Color trace(const gk::Ray& ray, int depth){
  gk::Hit hit(ray);
  gk::Color color;
  
  // calculer l'intersection du rayon avec les triangles
  m_bsp.intersect(ray,hit);
  //        intersect(ray, hit);
  
  // si une intersection existe...
  if(hit.object_id != -1){
    // calculer l'energie reflechie par le point vers la camera
    gk::Point p= ray(hit.t);
    //	gk::Normal normal= mesh->pntriangle(hit.object_id).normal(hit.u, hit.v);
    gk::Normal normal= mesh->triangle(hit.object_id).normal();
    const gk::MeshMaterial& material= mesh->triangleMaterial(hit.object_id);
    gk::MeshMaterial mUsed(material);
    color = material.emission;

    // + eclairage direct
    //    color = color ;
    if(NB_DIRECT>0)
      color += direct(p, normal, mUsed) ;
    if(NB_INDIRECT>0)
      color += indirect(p, normal, mUsed, hit.object_id, depth);

    if(mUsed.ns>EPSILON_MIRROR)
      color += reflexion(ray,p,normal,mUsed,depth) * mUsed.ns/256.f;
    color *= 0.5f;
    color = Max(color,1);
    color.a = 1;
  }
  return color;
}

static void * threadInit (void * p_data)
{
  return NULL;
}
static void * fn_render (void * ptr)
{

  thdata *data;            
  data = (thdata *) ptr;
  int nb = data->thread_no ;

  //  int h=Y; 
  //  int tmp = h/NB_THREAD;

  //  std::cout<<"=>OK"<<nb<<" "<<tmp<<"x"<<render.image->width<<std::endl;

  // parcours tous les pixels de l'image
  // for(int y= tmp * (nb-1); y < tmp * nb; y++){
  //   for(int x= 0; x < render.image->width; x++){

  for(int y= 0; y <  render.image->height; y++){
    if((y+1)%NB_THREAD == nb){
      std::cout<<"Ligne "<<y+1<<"/"<<render.image->height<<" Thread("<<nb<<"/"<<NB_THREAD<<")"<<std::endl;
      for(int x= 0; x < render.image->width; x++){
      // generer le rayon pour le pixel (x,y) dans le repere de l'image
      gk::Point origine(x +.5f, y + .5f, 0.f);    // sur le plan near
      gk::Point extremite(x +.5f, y + .5f, 1.f);    // sur le plan far
           
      // transformer le rayon dans le repere de la scene
      gk::Point o= render.vpv.inverse(origine); // transformation de origine, par exemple vpv(origine) ou vpv.inverse(origine)
      gk::Point e= render.vpv.inverse(extremite); // transformation de extremite
            
      gk::Ray ray(o, e);      
      render.image->setPixel(x, y, trace(ray, 0));
    }
  }
  }
  std::cout<<"=>END"<<nb<<std::endl;  
  return NULL;
}


int main( )
{
  int i = 0;
  int ret = pthread_create (
			    & render.thread_render, NULL,threadInit, NULL
			    );

  if(!ret){
    std::srand ( unsigned ( std::time(0) ) );

    // charger un objet
    //mesh= gk::MeshIO::readOBJ("hairball.obj");
    //mesh= gk::MeshIO::readOBJ("geometry.obj");
    //mesh= gk::MeshIO::readOBJ("secondary.obj");
     mesh= gk::MeshIO::readOBJ("emission.obj");
    gk::MeshIO::readMTL("materials.mtl",materials);
    
    //mesh= gk::MeshIO::readOBJ("shadow.obj");
    if(mesh == NULL)
      return 1;
    gk::MeshIO::buildNormals(mesh);   

    build_sources(mesh);        // recupere les sources de lumiere
    build_triangles(mesh);      // recupere les triangles

    std::sort (sources.begin(), sources.end(), compareSources);
    for(unsigned int i = 0; i<sources.size();i++)
      std::cout<<"source["<<i<<"]=>"<<sources[i].triangle.area()<<std::endl;

    m_bsp = BSP(triangles,0);

    render.c=new gk::Color[X*Y];    
    // creer une image resultat
    gk::Image *image= gk::createImage(X, Y);
    render.image = image;
    // definir les transformations
    gk::Transform model;
   
    /*
      geometry
      translate x y z
      -221.766296 232.837692 575.962341 
      rotate y x
      -378.000000 -7.000000
    */
    
    gk::Transform view= (gk::Translate( gk::Vector(-221.f, 232.f, 575.f) ) * gk::RotateX(-7.f) * gk::RotateY(-378.f)).inverse();
    gk::Transform projection= gk::Perspective(50.f, 1.f, 1.f, 1000.f);  // projection perspective
    gk::Transform viewport= gk::Viewport(image->width, image->height);      // transformation adaptee a la resolution de l'image resultat
    
    // compose les transformations utiles
    gk::Transform vpv= viewport * projection * view;
    render.vpv = vpv;
    thdata data[NB_THREAD]; 

    for (i = 0; i < NB_THREAD; i++){
    
      data[i].thread_no = i;
      ret = pthread_create (&render.thread_renders [i], NULL,fn_render, (void *) &data[i]);
    }

    for (i = 0; i < NB_THREAD; i++) {
      pthread_join (render.thread_renders [i], NULL);
    }


    pthread_join (render.thread_render, NULL);

    // enregistrer l'image
    gk::ImageIO::writeImage("render.hdr", image);
    delete image;
    delete []render.c;
  }   
    
  return 0;
}
